/*

Code extending the causal state reconstruction algorithm
technique exposed in "Automatic Filters for the Detection of Coherent
Structure in Spatiotemporal Systems". The algorithm below can
handle incremental estimation of the states as light cones
become available. Convergence is improved by running through the
state estimates in decreasing complexity. Also, states do really
handle all observations from their clustered past light cones.

The CausalStateAnalyzer class is generic and may easily be adapted.
Feed it light cones incrementally as you produce them, and it will
update the causal states and return you the best state estimate so far.

This code links to statistical testing routines that were developped
for the CSSR project (see Test.h and Test.cpp for proper credits).

Nicolas Brodu,
Released under GNU Public Licence v2 or above.
*/

#include <map>
#include <vector>

#include <assert.h>
#include <math.h>

#include "Test.h"

template<class PastLightCone, class FutureLightCone>
struct CausalStateAnalyzer {

    double significanceLevel;

    struct FutureDistribution : public std::map<FutureLightCone, int> {
        int total;

        FutureDistribution() : total(0) {}
        FutureDistribution(FutureLightCone flc) : total(0) {
            insert(flc);
        }

        int insert(FutureLightCone flc) {
            ++(*this)[flc];
            return ++total;
        }

        int remove(FutureLightCone flc) {
            typename std::map<FutureLightCone, int>::iterator flcIt = this->find(flc);
            assert(flcIt!=this->end()); // can't remove a future that was not inserted
            if (--flcIt->second == 0) this->erase(flcIt); // do not keep empty entries
            return --total; // may now be null
        }

        typedef typename std::map<FutureLightCone, int>::const_iterator const_iterator;

        void add(const FutureDistribution& distribution) {
            for (const_iterator it = distribution.begin(); it != distribution.end(); ++it) {
                (*this)[it->first] += it->second;
            }
            total += distribution.total;
        }

        void sub(const FutureDistribution& distribution) {
            for (const_iterator it = distribution.begin(); it != distribution.end(); ++it) {
                int newCount = (*this)[it->first] - it->second;
                assert(newCount>=0); // user can't remove a distribution that wasn't added
                if (newCount==0) erase(it->first);
                else (*this)[it->first] = newCount;
            }
            total -= distribution.total;
            assert(total>=0); // user can't remove a distribution that wasn't added
            // total maybe 0, but then, the cluster will be destroyed
        }

    };

    // Converts the distributions into a numerical format more appropriate for statistical testing
    struct DistributionTester : public Test {

        DistributionTester(bool isChiElseKS = true) : Test(isChiElseKS) {}

        // test if the 2 distribution match. Return value for significance testing
        double match(FutureDistribution dist1, FutureDistribution dist2) {
            std::vector<double> data1, data2;
            // create a bin for each distinct entry in either of the distributions
            // use the fact that std::map is sorted
            typename FutureDistribution::const_iterator it1 = dist1.begin();
            typename FutureDistribution::const_iterator it2 = dist2.begin();
            while (it1!=dist1.end() || it2!=dist2.end()) {
                if (it1->first<it2->first || it2==dist2.end()) {
                    data1.push_back(it1->second); ++it1;
                    data2.push_back(0);
                }
                else if (it2->first<it1->first || it1==dist1.end()) {
                    data1.push_back(0);
                    data2.push_back(it2->second); ++it2;
                }
                else {
                    data1.push_back(it1->second); ++it1;
                    data2.push_back(it2->second); ++it2;
                }
            }
            // call parent test on the distributions. Both vectors have the same size by construction
            return RunTest(data1, dist1.total, data2, dist2.total, data1.size());
        }
    };

    DistributionTester tester;

    // A cluster contains a an average future distribution
    // the mapping between past light cones to cluster could be done too
    struct Cluster {

        // The average distribution
        FutureDistribution distribution;
        // The number of times this cluster was observed
        int ninstances;

        // Individual pasts & their observed future distributions
        std::map<PastLightCone, FutureDistribution*> observedDistributions;

        // cluster complexity is -log(ninstances/totalInstances);
        // keep log(ninstances) since this is only modified when changing dist
        // then log(totalInstances) is computed only once for all clusters, and
        // a simple subtraction gives the complexity (instead of a log)
        double logNinstances;
        // ref back into vector for O(1) deletion
        int index;

        Cluster() : ninstances(0) {
        }

        void merge(PastLightCone plc, FutureDistribution* dist) {
            typename std::map<PastLightCone, FutureDistribution*>::iterator plcIt = observedDistributions.find(plc);
            assert(plcIt == observedDistributions.end());
            observedDistributions.insert(plcIt, make_pair(plc, dist));
            distribution.add(*dist);
            //observedDistributions[plc] = dist;
            ninstances += dist->total;
            logNinstances = log2(ninstances);
        }

        // Remove a plc from this cluster and return the corresponding flc distribution
        FutureDistribution* remove(PastLightCone plc) {
            typename std::map<PastLightCone, FutureDistribution*>::iterator plcIt = observedDistributions.find(plc);
            assert(plcIt != observedDistributions.end()); // cannot remove a past that was not inserted
            FutureDistribution* dist = plcIt->second;
            observedDistributions.erase(plcIt);
            distribution.sub(*dist);
            ninstances -= dist->total;
            assert(ninstances>=0); // and log = -inf doesn't matter for destroyed cluster
            logNinstances = log2(ninstances);
            return dist;
        }
    };

    typedef std::map<PastLightCone,Cluster*> ClusterMap;
    ClusterMap plcToCluster;

    int totalInstances;
    // subtract each cluster logNinstances from that to get that cluster complexity
    double logTotalInstances;

    // maintain random access to the existing clusters.
    // use pointers so they can be moved around without deep copy
    std::vector<Cluster*> clusters;

    // complexity map: especially useful to maintain min/max in sublinear time
    // indexed by ninstances as this is equivalent to indexing by complexity
    // and allows to keep the map when totalInstances changes
    // and also one can get logNinstances by looking into the cluster
    typedef std::multimap<int, Cluster*> ComplexityMap;
    ComplexityMap complexityMap;

    typedef int (FutureDistribution::*insert_or_remove_function)(FutureLightCone);

    template<insert_or_remove_function insert_or_remove, int deltaObs> Cluster* add_or_remove_Observation(PastLightCone plc, FutureLightCone flc) {
        // O(log(n_cluster)) look for a cluster for that plc
        typename ClusterMap::iterator clIt = plcToCluster.find(plc);

        FutureDistribution* observedDistribution;

        Cluster* oldCluster = 0;

        if (clIt != plcToCluster.end()) {
            bool found = false;
            // update complexityMap too: remove that cluster as its complexity will change anyway
            std::pair<typename ComplexityMap::iterator, typename ComplexityMap::iterator>
            eqRange = complexityMap.equal_range(clIt->second->ninstances);
            for (; eqRange.first != eqRange.second; ++eqRange.first) {
                if (eqRange.first->second == clIt->second) {
                    complexityMap.erase(eqRange.first);
                    found = true;
                    break;
                }
            }
            assert(found);

            // remove past from its cluster & get corresponding distribution
            observedDistribution = clIt->second->remove(plc);
            // destroy the cluster if it has become empty
            if (clIt->second->ninstances==0) {
                clusters[clIt->second->index] = clusters.back();
                clusters.back()->index = clIt->second->index;
                clusters.pop_back();
                delete clIt->second;
            }
            else oldCluster = clIt->second;
            // update that past observed distribution
            if ((observedDistribution->*insert_or_remove)(flc) == 0) {
                delete observedDistribution;
                observedDistribution = 0;
            }
        }
        // create a new distribution if this plc was not found, ensure it was found if we're removing an observation
        else {
            assert(deltaObs>0);
            observedDistribution = new FutureDistribution(flc);
        }

        Cluster* cluster = 0;

        // run through the clusters from max complexity to low, take first match.
        // Idea: if 2 clusters pass the stat. test, take the more complex one.
        // this forces high-complexity clusters to be really different hence enhance the contrast.
        // conversely the low-complexity are better merged as their elements get reevaluated after the initial phase.
        if (observedDistribution)
        for (typename ComplexityMap::iterator cit = complexityMap.begin(); cit!=complexityMap.end(); ++cit) {
            if (tester.match(*observedDistribution, cit->second->distribution) < significanceLevel) {
                cluster = cit->second;
                break;
            }
        }

        // no cluster found, create a new one if necessary
        if (cluster == 0) {
            if (observedDistribution) {
                cluster = new Cluster();
                cluster->index = clusters.size();
                clusters.push_back(cluster);
            }

            // new cluster is really new, can safely re-insert the distinct old one if it exists
            if (oldCluster) complexityMap.insert(make_pair(oldCluster->ninstances, oldCluster));
        }

        // reinsert old cluster without the distribution in the complexity map if it differs
        else if (oldCluster != cluster) {
            // and remove new cluster since its complexity will change
            bool found = false;
            std::pair<typename ComplexityMap::iterator, typename ComplexityMap::iterator>
            eqRange = complexityMap.equal_range(cluster->ninstances);
            for (; eqRange.first != eqRange.second; ++eqRange.first) {
                if (eqRange.first->second == cluster) {
                    complexityMap.erase(eqRange.first);
                    found = true;
                    break;
                }
            }
            assert(found);

            if (oldCluster) complexityMap.insert(make_pair(oldCluster->ninstances, oldCluster));
        }

        // update new cluster
        if (cluster!=0) {
            cluster->merge(plc, observedDistribution);

            if (clIt != plcToCluster.end()) clIt->second = cluster;
            else plcToCluster[plc] = cluster;

            // update complexity map for new cluster
            complexityMap.insert(make_pair(cluster->ninstances, cluster));
        }

        // update total & log
        totalInstances += deltaObs;
        logTotalInstances = log2(totalInstances);

        return cluster;
    }


    ~CausalStateAnalyzer() {
        for (unsigned int i=0; i<clusters.size(); ++i) delete clusters[i];
    }


// USER API BELOW

    CausalStateAnalyzer(double _significanceLevel, bool isChiElseKS = true) : significanceLevel(_significanceLevel), tester(isChiElseKS), totalInstances(0) {
    }

    // Add a new observed future light cone for a given past light cone
    // This will update the plc distibution and the clusters and their complexity
    // The (potentially new) cluster for that plc is returned.
    Cluster* addObservation(PastLightCone plc, FutureLightCone flc) {
        return add_or_remove_Observation<&FutureDistribution::insert,1>(plc,flc);
    }

    // Useful for systems changing over time or space. When "old" observations
    // have a different enough distribution from current observation (non-stationary system)
    // and should be discarded.
    // If there are still instances for that plc, the corresponding cluster is returned, otherwise 0.
    Cluster* removeObservation(PastLightCone plc, FutureLightCone flc) {
        return add_or_remove_Observation<&FutureDistribution::remove,-1>(plc,flc);
    }

    // get absolute cluster complexity
    double getComplexity(Cluster* cluster) {
        return logTotalInstances - cluster->logNinstances;
    }

    // scale complexity to 0..1 relatively to the other clusters
    double getScaledComplexity(Cluster* cluster) {
        double minLog = complexityMap.begin()->second->logNinstances;
        double maxLog = complexityMap.rbegin()->second->logNinstances;
        double scaleFactor = maxLog - minLog;
        if (scaleFactor!=0.0) scaleFactor = 1.0 / scaleFactor;
        return (maxLog - cluster->logNinstances) * scaleFactor;
    }

};

